#include <stdio.h> /*
set -x
gcc -g -W -Wall -O2 -DWANT_DETECTION_MAIN detection.c -o detec
exit
*/
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <SDL.h>

#include "detection.h"
#include "re.h"
#include "xvmc_tests.h"

/* globals, the detected devices to use */
pci_device_t *pcidev_ = NULL;
char device_node_[256] = {0};

/* private */
static const char pcidevsfile_[] = "/proc/bus/pci/devices";
static pci_device_t *pcilist_ = NULL;

static const char pcidevsname[] = "/proc/bus/pci/%02x/%02x.%x";

#ifdef WANT_DETECTION_MAIN

/* just a simple test of parsing functions. */
int main(void)
{
	if(create_pcilist() != 0)
		ERR_PRINT("List parsing failed, no nvidia devices present?\n");
	print_pcilist();
	destroy_pcilist();
	return 0;
}

#endif

/* Scan a single line from pcidevsfile_.
 * Return EOF on end-of-file or parse error.
 * Return 0 on line to be skipped.
 */
static int parse_pci_line(pci_device_t *p, FILE *fp)
{
	int i, ret;
	char buf[512];
	char filename[256],config_space[64];
	FILE *fc;
	
	ret = fscanf(fp, "%2x%2x %4hx%4hx %x",
		&p->bus_n, &p->devfn, &p->vendor, &p->device, &p->irq);
	if( ret != 5)
		return EOF;
	
	for(i=0; i<PCI_NUM_RESOURCES; ++i)
	{
		ret = fscanf(fp, " %lx", &p->resource[i].start);
		if( ret != 1 )
			return EOF;
	}
	for(i=0; i<PCI_NUM_RESOURCES; ++i)
	{
		ret = fscanf(fp, " %lx", &p->resource[i].len);
		if( ret != 1 )
			return EOF;
	}
	
	i=0;
	while( ((ret=fgetc(fp)) != EOF) && (i<511) )
	{
		if(ret == '\n')
			break;
		if(isspace(ret))
			continue;
		buf[i++] = ret;
	}
	buf[i] = 0;
	p->driver = NULL;

	/* Now load configuration space for this device */
	snprintf(filename, sizeof(filename), pcidevsname, p->bus_n,
		(p->devfn>>3) & 0x1f, p->devfn & 7);
	//fprintf(stderr, "Trying %s\n", filename);
	fc = fopen(filename, "rb");
	if (!fc) {
		return 0;		
	}

	fread(&config_space, 1,64,fc);
	fclose(fc);

	/* check if Display controller */
	if (config_space[0x0b] != 0x03) {
		//fprintf(stderr, "Not display controller\n");
		return 0;
	}
	/* check if VGA compatible controller */
	if (config_space[0x0a] != 0x00) {
		//fprintf(stderr, "Not VGA compatible controller\n");
		return 0;
	}
	/* check if VGA */
	if (config_space[0x09] != 0x00) {
		//fprintf(stderr, "Not VGA\n");
		return 0;
	}
	
	/* check if Nvidia vendor */
	if(p->vendor != 0x10de)
		return 0;

	/* accept deviced managed by driver "nvidia" */
	if(strcmp(buf, "nvidia"))
		return 0;
	
	printf("device\t%s\n", filename);

	p->driver = strdup(buf);
	return 1;
}

/* Create a linked list of pci_device_t by parsing pcidevsfile_.
 * Return 0 on success, 1 on failure.
 * destroy_pcilist() must be called for cleanup.
 */
int create_pcilist()
{
	int ret;
	pci_device_t **cur = &pcilist_;
	
	destroy_pcilist();
	
	FILE *fp = fopen(pcidevsfile_, "r");
	if(!fp)
	{
		int err = errno;
		ERR_PRINT("Opening pci device list '%s' failed,"
			  " because: %s\n", pcidevsfile_, strerror(err));
		return 1;
	}
	
	*cur = calloc(1, sizeof(pci_device_t));
	while(1)
	{
		ret = parse_pci_line(*cur, fp);
		if( ret == 0 )
			continue;
		if( ret == EOF )
		{
			free(*cur);
			*cur = NULL;
			break;
		}
		cur = &(*cur)->next;
		*cur = calloc(1, sizeof(pci_device_t));
	}
	
	fclose(fp);
	if(!pcilist_)
		return 1;
	return 0;
}

void print_pcilist()
{
	pci_device_t *p = pcilist_;
	int i;
	while(p)
	{
		printf("%02x%02x\t%04x%04x\t%x", p->bus_n, p->devfn,
			p->vendor, p->device, p->irq);
		for(i=0; i<PCI_NUM_RESOURCES; ++i)
			printf("\t%16lx", p->resource[i].start);
		for(i=0; i<PCI_NUM_RESOURCES; ++i)
			printf("\t%16lx", p->resource[i].len);
		printf("\t%s\n", p->driver);

		p = p->next;
	}
}

void destroy_pcilist()
{
	while(pcilist_)
	{
		pci_device_t *p = pcilist_->next;
		if(pcilist_->driver)
			free(pcilist_->driver);
		free(pcilist_);
		pcilist_ = p;
	}
}


/* Return pointer to a pci device with searched for resource.
 * The pointer is freed by destroy_pcilist().
 */
pci_device_t *find_pcidev_by_addr(unsigned long addr, unsigned long size)
{
	pci_device_t *p = pcilist_;
	while(p)
	{
		int i;
		for(i=0; i<PCI_NUM_RESOURCES; ++i)
		{
			unsigned long end = p->resource[i].start +
			                    p->resource[i].len;
			if(addr < p->resource[i].start)
				continue;
			if(addr >= end)
				continue;
			if( (addr+size) <= end )
				return p;
			ERR_PRINT("Strange! 0x%lx bytes at 0x%08lx goes "
			          "beyond a PCI resource.\n", size, addr);
		}
		p = p->next;
	}
	return NULL;
}

/* Return pointer to the n'th pci device that passed
 * the criteria in parse_pci_line().
 * The pointer is freed by destroy_pcilist().
 */
pci_device_t *nth_pcidev(unsigned int n)
{
	pci_device_t *p = pcilist_;
	while(n && p)
	{
		p = p->next;
		n--;
	}
	return p;
}


/* Interpret a string as a line from /proc/<pid>/maps,
 * parse contents and store it into mapping entry *map.
 * Return 0 on success.
 */
int parse_mapping(const char *line, mapping *map)
{
	unsigned long vm_start, vm_end, phys, inode;
	char prot[4];
	unsigned char major, minor;
	char devfile[256] = {0};
	int ret;
		
	ret = sscanf(line, "%lx-%lx %4c %lx %2hhx:%2hhx %lu %253s",
		&vm_start, &vm_end,
		prot, &phys, &major, &minor, &inode, devfile);

	/* reading devfile is allowed to fail */
	if(ret < 7)
		return 1;
	
	map->start = (unsigned int*)vm_start;
	map->phys = phys;
	map->size = vm_end - vm_start;
	map->before = NULL;
	map->after = NULL;
	map->type = 0;
	map->devfile = strdup(devfile);
	
	return 0;
}

/* Release all resources a mapping entry has allocated. */
void destroy_mapping(mapping *map)
{
	if(map->devfile)
		free(map->devfile);
	if(map->before)
		free(map->before);
	if(map->after)
		free(map->after);
	memset(map, 0, sizeof(mapping));
}

/* Check if the mapping entry is relevant:
 * is the associated file /dev/nvidia##
 */
int is_wanted_mapping(mapping *map)
{
	if( strncmp(map->devfile, "/dev/nvidia", 11) )
		return 0;
	if( strlen(map->devfile) > 13 )
		return 0;
	return 1;
}


/* Find the PCI device and the device node of used card.
 * Return 0 on success */
static int find_pci_device()
{
	FILE *fp;
	mapping map;
	char s[1024];
	
	fp = fopen("/proc/self/maps", "r");
	if(!fp)
	{
		ERR_PRINT("Opening /proc/self/maps failed?! (%s)\n",
			strerror(errno));
		return 1;
	}

	while(fgets(s, 1024, fp))
	{
		pci_device_t *dev;
		
		if(parse_mapping(s, &map))
		{
			ERR_PRINT("Failed to parse line: %s\n", s);
			continue;
		}
		/*ERR_PRINT("trying phys 0x%lx, '%s'\n", map.phys,
		        map.devfile);*/
		if(!is_wanted_mapping(&map))
		{
		/*	ERR_PRINT("Discarding line: %s\n", s); */
			destroy_mapping(&map);
			continue;
		}
		
		dev = find_pcidev_by_addr(map.phys, map.size);
		
		if(dev)
		{
			printf("# Found match, mapping '%s' 0x%lx "
			       "and pci %02x:%02x device %04x:%04x (%s)\n",
			       map.devfile, map.phys, dev->bus_n, dev->devfn,
			       dev->vendor, dev->device, dev->driver);
			
			if(!pcidev_)
			{
				pcidev_ = dev;
				strncpy(device_node_, map.devfile, 256);
				device_node_[255] = 0;
			}
			else if(pcidev_ != dev)
			{
				printf("# Strange: memory maps match "
				        "several PCI devices.\n");
			}
			else
				printf("#    - the first match confirmed.\n");
		}
		
		destroy_mapping(&map);
	}
	
	fclose(fp);
	
	if(!pcidev_)
		return 2;
	
	return 0;
}

/* The main device detection routine.
 * Open an SDL OpenGL window, and reads /proc/self/maps and
 * /proc/bus/pci/devices to determine the right devices.
 * Then destroys the window and releases SDL.
 */
int detect_devices(int opengl, int xvmc)
{
	SDL_Surface *surf;
	int ret;

	if(create_pcilist() != 0)
	{
		ERR_PRINT(
			"create_pcilist() failed.\n"
			"This means your Nvidia card was not detected.\n"
			"You should have the proprietary driver installed and "
			"running\nproperly before executing renouveau.\n"
			);
		return 1;
	}

	setup_SDL(opengl);
	surf = SDL_SetVideoMode(512, 512, 0,
				SDL_NOFRAME | (opengl ? SDL_OPENGL : 0));
	if(!surf)
	{
		ERR_PRINT(
			"Creating probe window failed.\n"
			"We tried to create a window by using SDL.\n"
			"Our OpenGL tests require a stencil buffer which"
			"usually\nis not available with 16-bit desktop color"
			"depth. If you are\nusing 16-bit colors, please try"
			"24-bit instead.\n"
			);
		return 1;
	}

	struct xvmc_state_t state;
	int xvmc_init = 0;
	if (xvmc) {
		if (init_xvmc_state(&state, 16, 16) < 0) {
			ERR_PRINT("Initializing XvMC failed.\n");
		} else {
			xvmc_init = 1;
			//init_xvmc_window(&state);
		}
	}

	/* This will set pcidev_ and device_node_. */
	ret = find_pci_device();

	if (xvmc_init)
		free_xvmc_state(&state);
	
	SDL_Quit();

	if(ret != 0)
	{
		ERR_PRINT("Device detection failed.\n");
		return 1;
	}
	
	return 0;
}


/* What is the right line for AGP aperture in /proc/iomem?
 * 'level' is the indentation level of a line
 * 'name' is the string after iomem addresses and colon
 * Returns 1 if accepted, otherwise 0.
 */
static int is_agp_iomem(int level, char *name)
{
	if(level > 0)
		return 0;
	
	if(!strcmp(name, "0000:00:00.0"))
		return 1;
	if(!strcmp(name, "GART"))
		return 1;
	return 0;
}

/* Detect the AGP aperture from /proc/iomem, with help
 * of is_agp_iomem().
 * Return zero, if aperture was found and recorded into *res.
 */
int find_agp_aperture(pci_resource *res)
{
	char buf[1024];
	FILE *fp;
	
	fp = fopen("/proc/iomem", "r");
	if(!fp)
	{
		ERR_PRINT("fopen: %s\n", strerror(errno));
		return 1;
	}
	
	while(fgets(buf, 1024, fp))
	{
		int level = 0;
		char *p = buf;
		while(*p && isspace(*p))
		{
			level++;
			p++;
		}
		
		unsigned long start, end;
		int n, ret;
		ret = sscanf(p, "%lx-%lx : %n", &start, &end, &n);
		if(ret < 2)
		{
			ERR_PRINT("corrupted line: %s\n", buf);
			continue;
		}
		p += n;
		n = strlen(p);
		if(p[n-1] == '\n')
			p[n-1] = '\0';
		
		if(is_agp_iomem(level, p))
		{
			res->start = start;
			res->len = end - start +1;
			fclose(fp);
			return 0;
		}
	}
	fclose(fp);
	return 2;
}
